import Flutter
import AVFoundation


enum Methods: String {
  case start
  case stop
}

enum Channels: String {
  case inputChannelName = "com.wearemobilefirst.audio_io.inputAudio";
  case outputChannelName = "com.wearemobilefirst.audio_io.outputAudio";
  case methodChannelName = "com.wearemobilefirst.audio_io";
}

public class SwiftAudioIoPlugin: NSObject, FlutterPlugin, FlutterStreamHandler {

    let engine = AVAudioEngine()
    var converter = AVAudioMixerNode()

    var _inputEventChannel: FlutterEventChannel?
    var _outputEventChannel: FlutterEventChannel? 
    var _inputEventSink: FlutterEventSink?

   // let buffer = RingBuffer<Float32>(48000);

    private lazy var sourceNode = AVAudioSourceNode { (_, _, frameCount, audioBufferList) -> OSStatus in
        let ablPointer = UnsafeMutableAudioBufferListPointer(audioBufferList)
        for frame in 0..<Int(frameCount) {
            for buffer in ablPointer {
               // let buf: UnsafeMutableBufferPointer<Float> = UnsafeMutableBufferPointer(buffer)
               // buf[frame] = self.buffer.read() ?? 0
            }
        }
        return noErr
    }
    
    private lazy var sinkNode = AVAudioSinkNode() { (timeStamp, frames, audioBufferList) ->
           OSStatus in
             let ptr = audioBufferList.pointee.mBuffers.mData?.assumingMemoryBound(to: Float.self)
             var monoSamples = [Float]()
             monoSamples.append(contentsOf: UnsafeBufferPointer(start: ptr, count: Int(frames)))
              self._inputEventSink?(monoSamples)
             return noErr
         }

  public static func register(with registrar: FlutterPluginRegistrar) {
    let channel = FlutterMethodChannel(name: Channels.methodChannelName.rawValue, binaryMessenger: registrar.messenger())
    let instance = SwiftAudioIoPlugin()
    registrar.addMethodCallDelegate(instance, channel: channel)
    instance._inputEventChannel = FlutterEventChannel(name: Channels.inputChannelName.rawValue , binaryMessenger: registrar.messenger())
    instance._inputEventChannel?.setStreamHandler(instance)                                                                 
    instance._outputEventChannel = FlutterEventChannel(name: Channels.outputChannelName.rawValue , binaryMessenger: registrar.messenger())        
                                                      
  }

  public func onListen(withArguments arguments: Any?,
                       eventSink: @escaping FlutterEventSink) -> FlutterError? {
        _inputEventSink = eventSink; 
    return nil            
  }

  public func onCancel(withArguments arguments: Any?) -> FlutterError? {
    _inputEventSink = nil
    return nil
  }

  public func handle(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
    print(call)
    switch call.method {
    case Methods.start.rawValue :
      self.start()
      result(nil)
      break;
    case Methods.stop.rawValue:
      self.stop()
      result(nil)
      break;
    default:
      result(FlutterMethodNotImplemented)
    }

  }
                                                                                                                                                                                      
  public func stop() {
     engine.stop()
  }

  public func start() {
    // Setup AVAudioSession
        do {
            try AVAudioSession.sharedInstance().setCategory(AVAudioSession.Category.playAndRecord)

            let ioBufferDuration = 128.0 / 44100.0

            try AVAudioSession.sharedInstance().setPreferredIOBufferDuration(ioBufferDuration)

        } catch {
            assertionFailure("AVAudioSession setup error: \(error)")
        }
        converter.outputVolume = 1.0
        // Setup engine and node instances
        let input = engine.inputNode
        let output = engine.mainMixerNode
        let inputFormat = input.outputFormat(forBus: 0)
        let outputFormat = output.inputFormat(forBus: 0)
        print(inputFormat)
        print(outputFormat)

        // Connect nodes
        engine.attach(converter)
        engine.attach(sinkNode)
        engine.attach(sourceNode)
        engine.connect(input, to: sinkNode, format: nil)
        engine.connect(sourceNode, to: converter, format: inputFormat)
        engine.connect(converter, to: output, format: outputFormat)

        // Start engine
        do {
            try engine.start()
        } catch {
            assertionFailure("AVAudioEngine start error: \(error)")
        }
  }
}

